/*
 *  This file is part of the Xen Crashdump Analyser.
 *
 *  The Xen Crashdump Analyser is free software: you can redistribute
 *  it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation, either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  The Xen Crashdump Analyser is distributed in the hope that it will
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Xen Crashdump Analyser.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *  Copyright (c) 2023 Citrix Inc.
 */

/**
 * @file include/prb5x.hpp
 * @author Per Bilse
 */

#include "types.hpp"
#include "memory.hpp"
#include "coreinfo.hpp"
#include "abstract/pagetable.hpp"

/*
 * This directive insists on static initialisers for all members;
 * that doesn't make much sense in the context here, but for the
 * sake of conformity, have it your way.
#pragma GCC diagnostic ignored "-Weffc++"
 */

/**
 * For a 5.10 or newer kernel, prb_size and prb_offset pass on architecture
 * specific sizeof() and offsetof() values (as read from the ELF note
 * section) to more generic code.  They contain a few superfluous members
 * that may prove useful later.  See note in prb5x.cpp for pointers to
 * further information.
 */

struct prb_size {
    long atomic_long_t;
    long dev_printk_info;
    long prb_data_blk_lpos;
    long prb_data_ring;
    long prb_desc;
    long prb_desc_ring;
    long printk_info;
    long printk_ringbuffer;
};

/**
 * See note for prb_size
 */

struct prb_offset {
    struct {
        long counter;
    } atomic_long_t;
    struct {
        long device;
        long subsystem;
    } dev_printk_info;
    struct {
        long begin;
        long next;
    } prb_data_blk_lpos;
    struct {
        long data;
        long head_lpos;
        long size_bits;
        long tail_lpos;
    } prb_data_ring;
    struct {
        long count_bits;
        long descs;
        long head_id;
        long infos;
        long tail_id;
    } prb_desc_ring;
    struct {
        long state_var;
        long text_blk_lpos;
    } prb_desc;
    struct {
        long caller_id;
        long dev_info;
        long seq;
        long text_len;
        long level;
        long ts_nsec;
    } printk_info;
    struct {
        long desc_ring;
        long fail;
        long text_data_ring;
    } printk_ringbuffer;
};

/**
 * Container and constructor for prb_size and prb_offset.
 */

struct PrbMetrics {
    struct prb_size size;
    struct prb_offset offset;
    /**
     * Constructor
     * @param info Coreinfo object containing size and offset data
     * @throws infolookup if symbol not found in core info section
     */
    PrbMetrics(const CoreInfo & info);
};

/**
 * Final computed data required to access the printk ringbuffer.
 * See note in prb5x.cpp for more details and pointers to information.
 */

struct PrbMap {
    char *prb;
    char *desc_ring;
    unsigned long desc_ring_count;
    char *descs;
    char *infos;
    char *text_data_ring;
    unsigned long text_data_ring_size;
    char *text_data;
    uint64_t head, tail;

    /// Constructor.
    PrbMap(void);

    /// Destructor.
    ~PrbMap(void);

    /**
     * Opposite of setup, release dynamically allocated members.
     */
    void teardown(void);

    /**
     * Initialise all members.
     * @param prb_addr Address of relevant 'prb' struct in core file.
     * @param m Size and offset data for struct members.
     * @param pt Pagetable for address lookup.
     * @throws std::bad_alloc from 'new'.
     */
    void setup(const uint64_t prb_addr,
               const PrbMetrics & m,
               const Abstract::PageTable & pt);

private:
    // @cond EXCLUDE
    PrbMap(const PrbMap&);
    PrbMap operator=(const PrbMap&);
    // @endcond
};
