/*
 *  This file is part of the Xen Crashdump Analyser.
 *
 *  The Xen Crashdump Analyser is free software: you can redistribute
 *  it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation, either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  The Xen Crashdump Analyser is distributed in the hope that it will
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Xen Crashdump Analyser.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *  Copyright (c) 2023 Citrix Inc.
 */

#include "prb5x.hpp"
#include "exceptions.hpp"
#include "util/log.hpp"
#include "util/macros.hpp"

/**
 * @file src/prb5x.cpp
 * @brief Machinations required to access a vmcore printk
 * ringbuffer ("prb") for kernels 5.10 and newer.
 * @author Per Bilse
 */

static uint64_t intval(const char *sym, const CoreInfo & info)
{
    uint64_t res;

    if ( ! info.lookup_key_dec_u64(sym, res) )
        throw infolookup(sym);
    return res;
}

/*
 * Initialise sizeof() and offsetof() values for kernel ringbuffer
 * structs from vmcore info/note section (see notes below).
 */

PrbMetrics::PrbMetrics(const CoreInfo & info) :
    size(), offset()
{
#define INIT_SIZE(x) (size.x = intval("SIZE(" #x ")", info))
#define INIT_OFFSET(x) (offset.x = intval("OFFSET(" #x ")", info))
    INIT_SIZE(atomic_long_t);
    INIT_SIZE(dev_printk_info);
    INIT_SIZE(prb_data_blk_lpos);
    INIT_SIZE(prb_data_ring);
    INIT_SIZE(prb_desc);
    INIT_SIZE(prb_desc_ring);
    INIT_SIZE(printk_info);
    INIT_SIZE(printk_ringbuffer);
    INIT_OFFSET(atomic_long_t.counter);
    INIT_OFFSET(dev_printk_info.device);
    INIT_OFFSET(dev_printk_info.subsystem);
    INIT_OFFSET(prb_data_blk_lpos.begin);
    INIT_OFFSET(prb_data_blk_lpos.next);
    INIT_OFFSET(prb_data_ring.data);
    INIT_OFFSET(prb_data_ring.head_lpos);
    INIT_OFFSET(prb_data_ring.size_bits);
    INIT_OFFSET(prb_data_ring.tail_lpos);
    INIT_OFFSET(prb_desc_ring.count_bits);
    INIT_OFFSET(prb_desc_ring.descs);
    INIT_OFFSET(prb_desc_ring.head_id);
    INIT_OFFSET(prb_desc_ring.infos);
    INIT_OFFSET(prb_desc_ring.tail_id);
    INIT_OFFSET(prb_desc.state_var);
    INIT_OFFSET(prb_desc.text_blk_lpos);
    INIT_OFFSET(printk_info.caller_id);
    INIT_OFFSET(printk_info.dev_info);
    INIT_OFFSET(printk_info.seq);
    INIT_OFFSET(printk_info.text_len);
    INIT_OFFSET(printk_info.ts_nsec);
    INIT_OFFSET(printk_ringbuffer.desc_ring);
    INIT_OFFSET(printk_ringbuffer.fail);
    INIT_OFFSET(printk_ringbuffer.text_data_ring);

    /* As of kernel 6.1.34, this is missing from the vmcore info section */
    try
    {
        INIT_OFFSET(printk_info.level);
    }
    catch ( const CommonError & )
    {
        LOG_INFO("Offset for printk ringbuffer log level byte not found, "
                 "default to 19\n");
        offset.printk_info.level = 19;
    }
#undef INIT_SIZE
#undef INIT_OFFSET
}

/*
 * Kernel 5.10 introduced a lockless printk ringbuffer, which required major
 * changes to implementation details.  It also expanded the info/notes
 * section in the vmcore file with all sizeof() and offsetof() data needed to
 * access the associated data structures in the vmcore file independently of
 * any C declarations or other information.  We store that information
 * in structs 'size' and 'offset' in PrbMetrics (see above), and the final
 * computed access information in PrbMap.  All of this is pretty much a
 * set piece that follows from the information embedded in the core file.
 *
 * Wrt details about how this is put together, there's a brief overview here:
 * https://lkml.org/lkml/2020/7/9/551
 * and more in-depth information and pointers can be found here:
 * https://linuxreviews.org/Linux_5.10_Will_Have_A_Fully_Lock-Less_Ring_Buffer
 * Detailed insight into the machinations can be gleaned from:
 * kernel sources:
 *     kernel/printk/printk.[ch]
 *     kernel/printk/printk_ringbuffer.[ch]
 *     Documentation/admin-guide/kdump/vmcoreinfo.rst
 * 'crash' utility (which the code leans heavily on):
 *     defs.h
 *     printk.c
 */

PrbMap::PrbMap(void) :
    prb(NULL),
    desc_ring(NULL),
    desc_ring_count(0),
    descs(NULL),
    infos(NULL),
    text_data_ring(NULL),
    text_data_ring_size(0),
    text_data(NULL),
    head(0),
    tail(0)
{}

PrbMap::~PrbMap(void)
{
    teardown();
}

void PrbMap::teardown(void)
{
    SAFE_DELETE_ARRAY(prb);
    SAFE_DELETE_ARRAY(descs);
    SAFE_DELETE_ARRAY(infos);
    SAFE_DELETE_ARRAY(text_data);
}

void PrbMap::setup(uint64_t prb_addr,
                   const PrbMetrics & m,
                   const Abstract::PageTable & pt)
{
    const struct prb_offset & po(m.offset);
    const struct prb_size & ps(m.size);
    uint64_t kaddr;
    unsigned size;

    size = ps.printk_ringbuffer;
    prb = new char[size];
    memory.read_block_vaddr(pt, prb_addr, prb, size);

    desc_ring = prb + po.printk_ringbuffer.desc_ring;
    size = *(uint32_t*)(desc_ring + po.prb_desc_ring.count_bits);
    desc_ring_count = 1UL << size;

    size = ps.prb_desc * desc_ring_count;
    descs = new char[size];
    kaddr = *(uint64_t*)(desc_ring + po.prb_desc_ring.descs);
    memory.read_block_vaddr(pt, kaddr, descs, size);

    size = ps.printk_info * desc_ring_count;
    infos = new char[size];
    kaddr = *(uint64_t*)(desc_ring + po.prb_desc_ring.infos);
    memory.read_block_vaddr(pt, kaddr, infos, size);

    text_data_ring = prb + po.printk_ringbuffer.text_data_ring;
    size = *(uint32_t*)(text_data_ring + po.prb_data_ring.size_bits);
    text_data_ring_size = 1UL << size;

    size = text_data_ring_size;
    text_data = new char[size];
    kaddr = *(uint64_t*)(text_data_ring + po.prb_data_ring.data);
    memory.read_block_vaddr(pt, kaddr, text_data, size);

    head = *(uint64_t*)(desc_ring + po.prb_desc_ring.head_id +
                        po.atomic_long_t.counter);
    tail = *(uint64_t*)(desc_ring + po.prb_desc_ring.tail_id +
                        po.atomic_long_t.counter);
}

/*
 * Local variables:
 * mode: C++
 * c-file-style: "BSD"
 * c-basic-offset: 4
 * tab-width: 4
 * indent-tabs-mode: nil
 * End:
 */
